@@ -25,6 +25,7 @@
 #include "fmtmacros.hpp"
 #include "language.hpp"
 
+#include <Depfile.hpp>
 #include <core/wincompat.hpp>
 #include <util/string.hpp>
 
@@ -35,14 +36,22 @@
 #include <cassert>
 
 using core::Statistic;
-using nonstd::nullopt;
-using nonstd::optional;
-using nonstd::string_view;
 
 namespace {
 
 enum class ColorDiagnostics : int8_t { never, automatic, always };
 
+// The dependency target in the dependency file is taken from the highest
+// priority source.
+enum class OutputDepOrigin : uint8_t {
+  // Not set
+  none = 0,
+  // From -MF target
+  mf = 1,
+  // From -Wp,-MD,target or -Wp,-MMD,target
+  wp = 2
+};
+
 struct ArgumentProcessingState
 {
   bool found_c_opt = false;
@@ -56,17 +65,17 @@ struct ArgumentProcessingState
   ColorDiagnostics color_diagnostics = ColorDiagnostics::automatic;
   bool found_directives_only = false;
   bool found_rewrite_includes = false;
-  nonstd::optional<std::string> found_xarch_arch;
+  std::optional<std::string> found_xarch_arch;
+  bool found_mf_opt = false;
+  bool found_wp_md_or_mmd_opt = false;
+  bool found_md_or_mmd_opt = false;
+  bool found_Wa_a_opt = false;
 
   std::string explicit_language;    // As specified with -x.
   std::string input_charset_option; // -finput-charset=...
 
-  // Is the dependency makefile name overridden with -MF?
-  bool dependency_filename_specified = false;
-
-  // Is the dependency target name implicitly specified using
-  // DEPENDENCIES_OUTPUT or SUNPRO_DEPENDENCIES?
-  bool dependency_implicit_target_specified = false;
+  // Is the dependency file set via -Wp,-M[M]D,target or -MFtarget?
+  OutputDepOrigin output_dep_origin = OutputDepOrigin::none;
 
   // Is the compiler being asked to output debug info on level 3?
   bool generating_debuginfo_level_3 = false;
@@ -98,6 +107,9 @@ struct ArgumentProcessingState
 
   // Whether to include the full command line in the hash.
   bool hash_full_command_line = false;
+
+  // Whether to include the actual CWD in the hash.
+  bool hash_actual_cwd = false;
 };
 
 bool
@@ -252,16 +264,17 @@ process_profiling_option(const Context& ctx,
   return true;
 }
 
-optional<Statistic>
-process_arg(const Context& ctx,
-            ArgsInfo& args_info,
-            Config& config,
-            Args& args,
-            size_t& args_index,
-            ArgumentProcessingState& state)
+// Returns std::nullopt if the option wasn't recognized, otherwise the error
+// code (with Statistic::none for "no error").
+std::optional<Statistic>
+process_option_arg(const Context& ctx,
+                   ArgsInfo& args_info,
+                   Config& config,
+                   Args& args,
+                   size_t& args_index,
+                   ArgumentProcessingState& state)
 {
   size_t& i = args_index;
-
   // The user knows best: just swallow the next arg.
   if (args[i] == "--ccache-skip") {
     i++;
@@ -270,7 +283,7 @@ process_arg(const Context& ctx,
       return Statistic::bad_compiler_arguments;
     }
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   bool changed_from_slash = false;
@@ -317,7 +330,7 @@ process_arg(const Context& ctx,
 
     args.replace(i, *file_args);
     i--;
-    return nullopt;
+    return Statistic::none;
   }
 
   // Handle cuda "-optf" and "--options-file" argument.
@@ -341,7 +354,7 @@ process_arg(const Context& ctx,
       args.insert(i + 1, *file_args);
     }
 
-    return nullopt;
+    return Statistic::none;
   }
 
   // These are always too hard.
@@ -374,7 +387,7 @@ process_arg(const Context& ctx,
     state.common_args.push_back(args[i]);
     state.common_args.push_back(args[i + 1]);
     ++i;
-    return nullopt;
+    return Statistic::none;
   }
 
   // Handle -arch options.
@@ -384,7 +397,7 @@ process_arg(const Context& ctx,
     if (args_info.arch_args.size() == 2) {
       config.set_run_second_cpp(true);
     }
-    return nullopt;
+    return Statistic::none;
   }
 
   // Some arguments that clang passes directly to cc1 (related to precompiled
@@ -404,11 +417,22 @@ process_arg(const Context& ctx,
     ++i;
   }
 
-  if (util::starts_with(args[i], "-Wa,")
-      && args[i].find('=') != std::string::npos) {
-    LOG("Assembler listing file (-Wa,...=file) is currently not supported: {}",
-        args[i]);
-    return Statistic::unsupported_compiler_option;
+  if (util::starts_with(args[i], "-Wa,")) {
+    for (const auto part : util::Tokenizer(&args[i][4], ",")) {
+      if (util::starts_with(part, "-a")) {
+        if (state.found_Wa_a_opt) {
+          LOG_RAW(
+            "Multiple assembler listing options (-Wa,-a) are not supported");
+          return Statistic::unsupported_compiler_option;
+        }
+        state.found_Wa_a_opt = true;
+
+        const auto eq_pos = part.find('=');
+        if (eq_pos != std::string_view::npos) {
+          args_info.output_al = part.substr(eq_pos + 1);
+        }
+      }
+    }
   }
 
   // Handle options that should not be passed to the preprocessor.
@@ -424,11 +448,11 @@ process_arg(const Context& ctx,
       state.compiler_only_args.push_back(args[i + 1]);
       ++i;
     }
-    return nullopt;
+    return Statistic::none;
   }
   if (compopt_prefix_affects_compiler_output(args[i])) {
     state.compiler_only_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   // Modules are handled on demand as necessary in the background, so there is
@@ -455,28 +479,27 @@ process_arg(const Context& ctx,
   // We must have -c.
   if (args[i] == "-c") {
     state.found_c_opt = true;
-    return nullopt;
+    return Statistic::none;
   }
 
   // MSVC -Fo with no space.
   if (util::starts_with(args[i], "-Fo") && config.is_compiler_group_msvc()) {
-    args_info.output_obj =
-      Util::make_relative_path(ctx, string_view(args[i]).substr(3));
-    return nullopt;
+    args_info.output_obj = args[i].substr(3);
+    return Statistic::none;
   }
 
   // when using nvcc with separable compilation, -dc implies -c
   if ((args[i] == "-dc" || args[i] == "--device-c")
       && config.compiler_type() == CompilerType::nvcc) {
     state.found_dc_opt = true;
-    return nullopt;
+    return Statistic::none;
   }
 
   // -S changes the default extension.
   if (args[i] == "-S") {
     state.common_args.push_back(args[i]);
     state.found_S_opt = true;
-    return nullopt;
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "-x")) {
@@ -486,7 +509,7 @@ process_arg(const Context& ctx,
       // language specification. (GCC's "-x" language argument is always
       // lowercase.)
       state.common_args.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     }
 
     // Special handling for -x: remember the last specified language before the
@@ -500,14 +523,14 @@ process_arg(const Context& ctx,
         state.explicit_language = args[i + 1];
       }
       i++;
-      return nullopt;
+      return Statistic::none;
     }
 
     DEBUG_ASSERT(args[i].length() >= 3);
     if (args_info.input_file.empty()) {
       state.explicit_language = args[i].substr(2);
     }
-    return nullopt;
+    return Statistic::none;
   }
 
   // We need to work out where the output was meant to go.
@@ -516,9 +539,9 @@ process_arg(const Context& ctx,
       LOG("Missing argument to {}", args[i]);
       return Statistic::bad_compiler_arguments;
     }
-    args_info.output_obj = Util::make_relative_path(ctx, args[i + 1]);
+    args_info.output_obj = args[i + 1];
     i++;
-    return nullopt;
+    return Statistic::none;
   }
 
   // Alternate form of -o with no space. Nvcc does not support this.
@@ -528,9 +551,8 @@ process_arg(const Context& ctx,
   if (util::starts_with(args[i], "-o")
       && config.compiler_type() != CompilerType::nvcc
       && config.compiler_type() != CompilerType::msvc) {
-    args_info.output_obj =
-      Util::make_relative_path(ctx, string_view(args[i]).substr(2));
-    return nullopt;
+    args_info.output_obj = args[i].substr(2);
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "-fdebug-prefix-map=")
@@ -538,7 +560,7 @@ process_arg(const Context& ctx,
     std::string map = args[i].substr(args[i].find('=') + 1);
     args_info.debug_prefix_maps.push_back(map);
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   // Debugging is handled specially, so that we know if we can strip line
@@ -550,12 +572,12 @@ process_arg(const Context& ctx,
       // Selection of DWARF format (-gdwarf or -gdwarf-<version>) enables
       // debug info on level 2.
       args_info.generating_debuginfo = true;
-      return nullopt;
+      return Statistic::none;
     }
 
     if (util::starts_with(args[i], "-gz")) {
       // -gz[=type] neither disables nor enables debug info.
-      return nullopt;
+      return Statistic::none;
     }
 
     char last_char = args[i].back();
@@ -572,21 +594,21 @@ process_arg(const Context& ctx,
         args_info.seen_split_dwarf = true;
       }
     }
-    return nullopt;
+    return Statistic::none;
   }
 
   // These options require special handling, because they behave differently
   // with gcc -E, when the output file is not specified.
   if ((args[i] == "-MD" || args[i] == "-MMD")
       && !config.is_compiler_group_msvc()) {
+    state.found_md_or_mmd_opt = true;
     args_info.generating_dependencies = true;
-    args_info.seen_MD_MMD = true;
     state.dep_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "-MF")) {
-    state.dependency_filename_specified = true;
+    state.found_mf_opt = true;
 
     std::string dep_file;
     bool separate_argument = (args[i].size() == 3);
@@ -602,7 +624,11 @@ process_arg(const Context& ctx,
       // -MFarg or -MF=arg (EDG-based compilers)
       dep_file = args[i].substr(args[i][3] == '=' ? 4 : 3);
     }
-    args_info.output_dep = Util::make_relative_path(ctx, dep_file);
+
+    if (state.output_dep_origin <= OutputDepOrigin::mf) {
+      state.output_dep_origin = OutputDepOrigin::mf;
+      args_info.output_dep = Util::make_relative_path(ctx, dep_file);
+    }
     // Keep the format of the args the same.
     if (separate_argument) {
       state.dep_args.push_back("-MF");
@@ -610,13 +636,14 @@ process_arg(const Context& ctx,
     } else {
       state.dep_args.push_back("-MF" + args_info.output_dep);
     }
-    return nullopt;
+    return Statistic::none;
   }
 
   if ((util::starts_with(args[i], "-MQ") || util::starts_with(args[i], "-MT"))
       && !config.is_compiler_group_msvc()) {
-    args_info.dependency_target_specified = true;
+    const bool is_mq = args[i][2] == 'Q';
 
+    std::string_view dep_target;
     if (args[i].size() == 3) {
       // -MQ arg or -MT arg
       if (i == args.size() - 1) {
@@ -624,16 +651,26 @@ process_arg(const Context& ctx,
         return Statistic::bad_compiler_arguments;
       }
       state.dep_args.push_back(args[i]);
-      std::string relpath = Util::make_relative_path(ctx, args[i + 1]);
-      state.dep_args.push_back(relpath);
+      state.dep_args.push_back(args[i + 1]);
+      dep_target = args[i + 1];
       i++;
     } else {
-      auto arg_opt = string_view(args[i]).substr(0, 3);
-      auto option = string_view(args[i]).substr(3);
-      auto relpath = Util::make_relative_path(ctx, option);
-      state.dep_args.push_back(FMT("{}{}", arg_opt, relpath));
+      // -MQarg or -MTarg
+      const std::string_view arg_view(args[i]);
+      const auto arg_opt = arg_view.substr(0, 3);
+      dep_target = arg_view.substr(3);
+      state.dep_args.push_back(FMT("{}{}", arg_opt, dep_target));
     }
-    return nullopt;
+
+    if (args_info.dependency_target) {
+      args_info.dependency_target->push_back(' ');
+    } else {
+      args_info.dependency_target = "";
+    }
+    *args_info.dependency_target +=
+      is_mq ? Depfile::escape_filename(dep_target) : dep_target;
+
+    return Statistic::none;
   }
 
   // MSVC -MD[d], -MT[d] and -LT[d] options are something different than GCC's
@@ -644,25 +681,31 @@ process_arg(const Context& ctx,
     // These affect compiler but also #define some things.
     state.cpp_args.push_back(args[i]);
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
+  }
+
+  if (args[i] == "-showIncludes") {
+    args_info.generating_includes = true;
+    state.dep_args.push_back(args[i]);
+    return Statistic::none;
   }
 
   if (args[i] == "-fprofile-arcs") {
     args_info.profile_arcs = true;
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-ftest-coverage") {
     args_info.generating_coverage = true;
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-fstack-usage") {
     args_info.generating_stackusage = true;
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   // -Zs is MSVC's -fsyntax-only equivalent
@@ -670,7 +713,7 @@ process_arg(const Context& ctx,
     args_info.expect_output_obj = false;
     state.compiler_only_args.push_back(args[i]);
     state.found_syntax_only = true;
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "--coverage"      // = -fprofile-arcs -ftest-coverage
@@ -678,7 +721,16 @@ process_arg(const Context& ctx,
     args_info.profile_arcs = true;
     args_info.generating_coverage = true;
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
+  }
+
+  if (args[i] == "-fprofile-abs-path") {
+    if (!config.sloppiness().is_enabled(core::Sloppy::gcno_cwd)) {
+      // -fprofile-abs-path makes the compiler include absolute paths based on
+      // the actual CWD in the .gcno file.
+      state.hash_actual_cwd = true;
+    }
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "-fprofile-")
@@ -689,20 +741,20 @@ process_arg(const Context& ctx,
       return Statistic::unsupported_compiler_option;
     }
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "-fsanitize-blacklist=")) {
     args_info.sanitize_blacklists.emplace_back(args[i].substr(21));
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "--sysroot=")) {
-    auto path = string_view(args[i]).substr(10);
+    auto path = std::string_view(args[i]).substr(10);
     auto relpath = Util::make_relative_path(ctx, path);
     state.common_args.push_back("--sysroot=" + relpath);
-    return nullopt;
+    return Statistic::none;
   }
 
   // Alternate form of specifying sysroot without =
@@ -715,7 +767,7 @@ process_arg(const Context& ctx,
     auto relpath = Util::make_relative_path(ctx, args[i + 1]);
     state.common_args.push_back(relpath);
     i++;
-    return nullopt;
+    return Statistic::none;
   }
 
   // Alternate form of specifying target without =
@@ -727,7 +779,7 @@ process_arg(const Context& ctx,
     state.common_args.push_back(args[i]);
     state.common_args.push_back(args[i + 1]);
     i++;
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-P" || args[i] == "-Wp,-P") {
@@ -736,7 +788,7 @@ process_arg(const Context& ctx,
     state.compiler_only_args.push_back(args[i]);
     LOG("{} used; not compiling preprocessed code", args[i]);
     config.set_run_second_cpp(true);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (util::starts_with(args[i], "-Wp,")) {
@@ -747,34 +799,37 @@ process_arg(const Context& ctx,
       return Statistic::unsupported_compiler_option;
     } else if (util::starts_with(args[i], "-Wp,-MD,")
                && args[i].find(',', 8) == std::string::npos) {
+      state.found_wp_md_or_mmd_opt = true;
       args_info.generating_dependencies = true;
-      state.dependency_filename_specified = true;
-      args_info.output_dep =
-        Util::make_relative_path(ctx, string_view(args[i]).substr(8));
+      if (state.output_dep_origin <= OutputDepOrigin::wp) {
+        state.output_dep_origin = OutputDepOrigin::wp;
+        args_info.output_dep = args[i].substr(8);
+      }
       state.dep_args.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     } else if (util::starts_with(args[i], "-Wp,-MMD,")
                && args[i].find(',', 9) == std::string::npos) {
+      state.found_wp_md_or_mmd_opt = true;
       args_info.generating_dependencies = true;
-      state.dependency_filename_specified = true;
-      args_info.output_dep =
-        Util::make_relative_path(ctx, string_view(args[i]).substr(9));
+      if (state.output_dep_origin <= OutputDepOrigin::wp) {
+        state.output_dep_origin = OutputDepOrigin::wp;
+        args_info.output_dep = args[i].substr(9);
+      }
       state.dep_args.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     } else if (util::starts_with(args[i], "-Wp,-D")
                && args[i].find(',', 6) == std::string::npos) {
       // Treat it like -D.
       state.cpp_args.push_back(args[i].substr(4));
-      return nullopt;
+      return Statistic::none;
     } else if (args[i] == "-Wp,-MP"
                || (args[i].size() > 8 && util::starts_with(args[i], "-Wp,-M")
                    && args[i][7] == ','
                    && (args[i][6] == 'F' || args[i][6] == 'Q'
                        || args[i][6] == 'T')
                    && args[i].find(',', 8) == std::string::npos)) {
-      // TODO: Make argument to MF/MQ/MT relative.
       state.dep_args.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     } else if (config.direct_mode()) {
       // -Wp, can be used to pass too hard options to the preprocessor.
       // Hence, disable direct mode.
@@ -784,18 +839,18 @@ process_arg(const Context& ctx,
 
     // Any other -Wp,* arguments are only relevant for the preprocessor.
     state.cpp_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-MP") {
     state.dep_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   // Input charset needs to be handled specially.
   if (util::starts_with(args[i], "-finput-charset=")) {
     state.input_charset_option = args[i];
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "--serialize-diagnostics") {
@@ -806,7 +861,7 @@ process_arg(const Context& ctx,
     args_info.generating_diagnostics = true;
     args_info.output_dia = Util::make_relative_path(ctx, args[i + 1]);
     i++;
-    return nullopt;
+    return Statistic::none;
   }
 
   if (config.compiler_type() == CompilerType::gcc) {
@@ -814,16 +869,16 @@ process_arg(const Context& ctx,
         || args[i] == "-fdiagnostics-color=always") {
       state.color_diagnostics = ColorDiagnostics::always;
       state.compiler_only_args_no_hash.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     } else if (args[i] == "-fno-diagnostics-color"
                || args[i] == "-fdiagnostics-color=never") {
       state.color_diagnostics = ColorDiagnostics::never;
       state.compiler_only_args_no_hash.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     } else if (args[i] == "-fdiagnostics-color=auto") {
       state.color_diagnostics = ColorDiagnostics::automatic;
       state.compiler_only_args_no_hash.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     }
   } else if (config.is_compiler_group_clang()) {
     // In the "-Xclang -fcolor-diagnostics" form, -Xclang is skipped and the
@@ -836,36 +891,36 @@ process_arg(const Context& ctx,
     if (args[i] == "-fcolor-diagnostics") {
       state.color_diagnostics = ColorDiagnostics::always;
       state.compiler_only_args_no_hash.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     } else if (args[i] == "-fno-color-diagnostics") {
       state.color_diagnostics = ColorDiagnostics::never;
       state.compiler_only_args_no_hash.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     }
   }
 
   // GCC
   if (args[i] == "-fdirectives-only") {
     state.found_directives_only = true;
-    return nullopt;
+    return Statistic::none;
   }
 
   // Clang
   if (args[i] == "-frewrite-includes") {
     state.found_rewrite_includes = true;
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-fno-pch-timestamp") {
     args_info.fno_pch_timestamp = true;
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-fpch-preprocess") {
     state.found_fpch_preprocess = true;
     state.common_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (config.sloppiness().is_enabled(core::Sloppy::clang_index_store)
@@ -877,7 +932,7 @@ process_arg(const Context& ctx,
     if (i <= args.size() - 1) {
       LOG("Skipping argument -index-store-path {}", args[i]);
     }
-    return nullopt;
+    return Statistic::none;
   }
 
   if (args[i] == "-frecord-gcc-switches") {
@@ -887,7 +942,7 @@ process_arg(const Context& ctx,
   // MSVC -u is something else than GCC -u, handle it specially.
   if (args[i] == "-u" && ctx.config.is_compiler_group_msvc()) {
     state.cpp_args.push_back(args[i]);
-    return nullopt;
+    return Statistic::none;
   }
 
   if (compopt_takes_path(args[i])) {
@@ -921,7 +976,7 @@ process_arg(const Context& ctx,
     dest_args.push_back(relpath);
 
     i += next;
-    return nullopt;
+    return Statistic::none;
   }
 
   // Detect PCH for options with concatenated path (relative or absolute).
@@ -946,15 +1001,15 @@ process_arg(const Context& ctx,
     if (path_pos) {
       const std::string option = args[i].substr(0, *path_pos);
       if (compopt_takes_concat_arg(option) && compopt_takes_path(option)) {
-        const auto relpath =
-          Util::make_relative_path(ctx, string_view(args[i]).substr(*path_pos));
+        const auto relpath = Util::make_relative_path(
+          ctx, std::string_view(args[i]).substr(*path_pos));
         std::string new_option = option + relpath;
         if (compopt_affects_cpp_output(option)) {
           state.cpp_args.push_back(new_option);
         } else {
           state.common_args.push_back(new_option);
         }
-        return nullopt;
+        return Statistic::none;
       }
     }
   }
@@ -975,7 +1030,12 @@ process_arg(const Context& ctx,
     }
 
     i++;
-    return nullopt;
+    return Statistic::none;
+  }
+
+  if (args[i] == "--") {
+    args_info.seen_double_dash = true;
+    return Statistic::none;
   }
 
   // Other options.
@@ -986,7 +1046,7 @@ process_arg(const Context& ctx,
     } else {
       state.common_args.push_back(args[i]);
     }
-    return nullopt;
+    return Statistic::none;
   }
 
   // It was not a known option.
@@ -994,6 +1054,26 @@ process_arg(const Context& ctx,
     args[i][0] = '/';
   }
 
+  return std::nullopt;
+}
+
+Statistic
+process_arg(const Context& ctx,
+            ArgsInfo& args_info,
+            Config& config,
+            Args& args,
+            size_t& args_index,
+            ArgumentProcessingState& state)
+{
+  const auto processed =
+    process_option_arg(ctx, args_info, config, args, args_index, state);
+  if (processed) {
+    const auto& error = *processed;
+    return error;
+  }
+
+  size_t& i = args_index;
+
   // If an argument isn't a plain file then assume its an option, not an input
   // file. This allows us to cope better with unusual compiler options.
   //
@@ -1004,7 +1084,7 @@ process_arg(const Context& ctx,
     if (!st || !st.is_regular()) {
       LOG("{} is not a regular file, not considering as input file", args[i]);
       state.common_args.push_back(args[i]);
-      return nullopt;
+      return Statistic::none;
     }
   }
 
@@ -1026,67 +1106,18 @@ process_arg(const Context& ctx,
   }
 
   // Rewrite to relative to increase hit rate.
+  args_info.orig_input_file = args[i];
   args_info.input_file = Util::make_relative_path(ctx, args[i]);
   args_info.normalized_input_file =
     Util::normalize_concrete_absolute_path(args_info.input_file);
 
-  return nullopt;
+  return Statistic::none;
 }
 
-void
-handle_dependency_environment_variables(Context& ctx,
-                                        ArgumentProcessingState& state)
+const char*
+get_default_object_file_extension(const Config& config)
 {
-  ArgsInfo& args_info = ctx.args_info;
-
-  // See <http://gcc.gnu.org/onlinedocs/cpp/Environment-Variables.html>.
-  // Contrary to what the documentation seems to imply the compiler still
-  // creates object files with these defined (confirmed with GCC 8.2.1), i.e.
-  // they work as -MMD/-MD, not -MM/-M. These environment variables do nothing
-  // on Clang.
-  const char* dependencies_env = getenv("DEPENDENCIES_OUTPUT");
-  bool using_sunpro_dependencies = false;
-  if (!dependencies_env) {
-    dependencies_env = getenv("SUNPRO_DEPENDENCIES");
-    using_sunpro_dependencies = true;
-  }
-  if (!dependencies_env) {
-    return;
-  }
-
-  args_info.generating_dependencies = true;
-  state.dependency_filename_specified = true;
-
-  auto dependencies = Util::split_into_views(dependencies_env, " ");
-
-  if (!dependencies.empty()) {
-    auto abspath_file = dependencies[0];
-    args_info.output_dep = Util::make_relative_path(ctx, abspath_file);
-  }
-
-  // Specifying target object is optional.
-  if (dependencies.size() > 1) {
-    // It's the "file target" form.
-    ctx.args_info.dependency_target_specified = true;
-    string_view abspath_obj = dependencies[1];
-    std::string relpath_obj = Util::make_relative_path(ctx, abspath_obj);
-    // Ensure that the compiler gets a relative path.
-    std::string relpath_both = FMT("{} {}", args_info.output_dep, relpath_obj);
-    if (using_sunpro_dependencies) {
-      Util::setenv("SUNPRO_DEPENDENCIES", relpath_both);
-    } else {
-      Util::setenv("DEPENDENCIES_OUTPUT", relpath_both);
-    }
-  } else {
-    // It's the "file" form.
-    state.dependency_implicit_target_specified = true;
-    // Ensure that the compiler gets a relative path.
-    if (using_sunpro_dependencies) {
-      Util::setenv("SUNPRO_DEPENDENCIES", args_info.output_dep);
-    } else {
-      Util::setenv("DEPENDENCIES_OUTPUT", args_info.output_dep);
-    }
-  }
+  return config.is_compiler_group_msvc() ? ".obj" : ".o";
 }
 
 } // namespace
@@ -1107,15 +1138,34 @@ process_args(Context& ctx)
 
   state.common_args.push_back(args[0]); // Compiler
 
-  optional<Statistic> argument_error;
+  std::optional<Statistic> argument_error;
   for (size_t i = 1; i < args.size(); i++) {
-    const auto error =
-      process_arg(ctx, ctx.args_info, ctx.config, args, i, state);
-    if (error && !argument_error) {
+    const auto error = process_arg(ctx, args_info, ctx.config, args, i, state);
+    if (error != Statistic::none && !argument_error) {
       argument_error = error;
     }
   }
 
+  // Bail out on too hard combinations of options.
+  if (state.found_mf_opt && state.found_wp_md_or_mmd_opt) {
+    // GCC and Clang behave differently when "-Wp,-M[M]D,wp.d" and "-MF mf.d"
+    // are used: GCC writes to wp.d but Clang writes to mf.d. We could
+    // potentially support this by behaving differently depending on the
+    // compiler type, but let's just bail out for now.
+    LOG_RAW("-Wp,-M[M]D in combination with -MF is not supported");
+    return Statistic::unsupported_compiler_option;
+  }
+  if (state.found_wp_md_or_mmd_opt && !args_info.output_obj.empty()
+      && !state.found_md_or_mmd_opt && !args_info.dependency_target) {
+    // GCC and Clang behave differently when "-Wp,-M[M]D,wp.d" is used with "-o"
+    // but with neither "-MMD" nor "-MT"/"-MQ": GCC uses a dependency target
+    // based on the source filename but Clang bases it on the output filename.
+    // We could potentially support by behaving differently depending on the
+    // compiler type, but let's just bail out for now.
+    LOG_RAW("-Wp,-M[M]D with -o without -MMD, -MQ or -MT is not supported");
+    return Statistic::unsupported_compiler_option;
+  }
+
   // Don't try to second guess the compiler's heuristics for stdout handling.
   if (args_info.output_obj == "-") {
     LOG_RAW("Output file is -");
@@ -1137,18 +1187,17 @@ process_args(Context& ctx)
   }
 
   if (output_obj_by_source && !args_info.input_file.empty()) {
-    string_view extension;
-    if (state.found_S_opt) {
-      extension = ".s";
-    } else if (!ctx.config.is_compiler_group_msvc()) {
-      extension = ".o";
-    } else {
-      extension = ".obj";
-    }
+    std::string_view extension =
+      state.found_S_opt ? ".s" : get_default_object_file_extension(ctx.config);
     args_info.output_obj +=
       Util::change_extension(Util::base_name(args_info.input_file), extension);
   }
 
+  args_info.orig_output_obj = args_info.output_obj;
+  args_info.output_obj = Util::make_relative_path(ctx, args_info.output_obj);
+
+  // Determine output dependency file.
+
   // On argument processing error, return now since we have determined
   // args_info.output_obj which is needed to determine the log filename in
   // CCACHE_DEBUG mode.
@@ -1172,8 +1221,6 @@ process_args(Context& ctx)
   }
 #endif
 
-  handle_dependency_environment_variables(ctx, state);
-
   if (args_info.input_file.empty()) {
     LOG_RAW("No input file found");
     return Statistic::no_input_file;
@@ -1213,7 +1260,9 @@ process_args(Context& ctx)
     || Util::is_precompiled_header(args_info.output_obj);
 
   if (args_info.output_is_precompiled_header && output_obj_by_source) {
-    args_info.output_obj = args_info.input_file + ".gch";
+    args_info.orig_output_obj = args_info.orig_input_file + ".gch";
+    args_info.output_obj =
+      Util::make_relative_path(ctx, args_info.orig_output_obj);
   }
 
   if (args_info.output_is_precompiled_header
@@ -1295,8 +1344,8 @@ process_args(Context& ctx)
   // Some options shouldn't be passed to the real compiler when it compiles
   // preprocessed code:
   //
-  // -finput-charset=XXX (otherwise conversion happens twice)
-  // -x XXX (otherwise the wrong language is selected)
+  // -finput-charset=CHARSET (otherwise conversion happens twice)
+  // -x CHARSET (otherwise the wrong language is selected)
   if (!state.input_charset_option.empty()) {
     state.cpp_args.push_back(state.input_charset_option);
   }
@@ -1315,7 +1364,7 @@ process_args(Context& ctx)
 
   // Since output is redirected, compilers will not color their output by
   // default, so force it explicitly.
-  nonstd::optional<std::string> diagnostics_color_arg;
+  std::optional<std::string> diagnostics_color_arg;
   if (config.is_compiler_group_clang()) {
     // Don't pass -fcolor-diagnostics when compiling assembler to avoid an
     // "argument unused during compilation" warning.
@@ -1330,29 +1379,29 @@ process_args(Context& ctx)
   }
 
   if (args_info.generating_dependencies) {
-    if (!state.dependency_filename_specified) {
-      auto default_depfile_name =
-        Util::change_extension(args_info.output_obj, ".d");
-      args_info.output_dep =
-        Util::make_relative_path(ctx, default_depfile_name);
+    if (state.output_dep_origin == OutputDepOrigin::none) {
+      args_info.output_dep = Util::change_extension(args_info.output_obj, ".d");
       if (!config.run_second_cpp()) {
         // If we're compiling preprocessed code we're sending dep_args to the
         // preprocessor so we need to use -MF to write to the correct .d file
         // location since the preprocessor doesn't know the final object path.
         state.dep_args.push_back("-MF");
-        state.dep_args.push_back(default_depfile_name);
+        state.dep_args.push_back(args_info.output_dep);
       }
     }
 
-    if (!ctx.args_info.dependency_target_specified
-        && !state.dependency_implicit_target_specified
-        && !config.run_second_cpp()) {
+    if (!args_info.dependency_target && !config.run_second_cpp()) {
       // If we're compiling preprocessed code we're sending dep_args to the
       // preprocessor so we need to use -MQ to get the correct target object
       // file in the .d file.
       state.dep_args.push_back("-MQ");
       state.dep_args.push_back(args_info.output_obj);
     }
+
+    if (!args_info.dependency_target) {
+      args_info.dependency_target =
+        Depfile::escape_filename(args_info.orig_output_obj);
+    }
   }
 
   if (args_info.generating_stackusage) {
@@ -1446,9 +1495,21 @@ process_args(Context& ctx)
     }
     if (ctx.config.depend_mode()) {
       // The compiler is invoked with the original arguments in the depend mode.
-      ctx.args_info.depend_extra_args.push_back(*diagnostics_color_arg);
+      args_info.depend_extra_args.push_back(*diagnostics_color_arg);
     }
   }
 
-  return {preprocessor_args, extra_args_to_hash, compiler_args};
+  if (ctx.config.depend_mode() && !args_info.generating_includes
+      && ctx.config.compiler_type() == CompilerType::msvc) {
+    ctx.auto_depend_mode = true;
+    args_info.generating_includes = true;
+    args_info.depend_extra_args.push_back("/showIncludes");
+  }
+
+  return {
+    preprocessor_args,
+    extra_args_to_hash,
+    compiler_args,
+    state.hash_actual_cwd,
+  };
 }
